﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using AbbyyLS.SmartCat.AppIntegrations.Client;
using AbbyyLS.SmartCat.AppIntegrations.WebApi;
using AbbyyLS.SmartCat.AppIntegrations.WebApi.Client;
using AbbyyLS.SmartCat.AppIntegrations.WebApi.Client.Logging;
using Microsoft.Build.Framework;
using Microsoft.Build.Utilities;
using MsBuildTask = Microsoft.Build.Utilities.Task;
using Task = System.Threading.Tasks.Task;


namespace Resto.Build.Extensions.SmartCat
{
    public abstract class SmartCatBaseTask : MsBuildTask
    {
        private static readonly TimeSpan DefaultClientRequestTimeout = TimeSpan.FromMinutes(5);

        private readonly Lazy<SmartCatApiClient> translationApiClient;

        protected SmartCatApiClient TranslationApiClient => translationApiClient.Value;

        /// <summary>
        /// Включение/отключение записи в лог http-запросов и ответов.
        /// </summary>
        public bool EnableRequestsLog { get; set; } = false;

        protected SmartCatBaseTask()
        {
            translationApiClient = new Lazy<SmartCatApiClient>(GetLocalizationServiceClient);
        }

        public override bool Execute()
        {
            try
            {
                ValidateParameters();

                InternalExecute();
            }
            catch (Exception ex)
            {
                Log.LogErrorFromException(ex, true, true, GetType().Assembly.Location);
            }

            return !Log.HasLoggedErrors;
        }

        protected abstract void ValidateParameters();

        protected abstract void InternalExecute();

        private SmartCatApiClient GetLocalizationServiceClient()
        {
            var config = new SmartCatApiConfig("https://smartcat.ai", "IIKO", "aIFmJQeU", (uint)DefaultClientRequestTimeout.TotalMilliseconds);
            return new SmartCatApiClient(config, EnableRequestsLog ? new DebugLogManager(s => Log.LogMessage(s)) : null);
        }

        protected async Task<ProjectModel> GetProject(Guid projectId)
        {
            if (projectId == Guid.Empty)
            {
                return null;
            }

            return await TranslationApiClient.Projects.GetProject(projectId);
        }

        protected IDictionary GetDocumentMetadata(DocumentModel model)
        {
            return new Dictionary<string, object>
            {
                { "Name", model.Name},
                { "CreationDate", model.CreationDate},
                { "Deadline", model.Deadline},
                { "DocumentDisassemblingStatus", model.DocumentDisassemblingStatus},
                { "PretranslateCompleted", model.PretranslateCompleted},
                { "SourceLanguage", model.SourceLanguage},
                { "Status", model.Status},
                { "TargetLanguage", model.TargetLanguage}
            };
        }

        protected IDictionary GetProjectMetadata(ProjectModel project)
        {
            return new Dictionary<string, object>
            {
                {"Name", project.Name},
                {"Description", project.Description},
                {"SourceLanguage", project.SourceLanguage},
                {"TargetLanguages", string.Join(";", project.TargetLanguages) },
                {"ClientId", project.ClientId },
                {"CreationDate", project.CreationDate},
                {"Deadline", project.Deadline},
                {"DomainId", project.DomainId},
                {"ModificationDate", project.ModificationDate},
                {"Status", project.Status},
                {"VendorAccountId", project.VendorAccountId}
                // {"WorkflowStages", project.WorkflowStages},
            };
        }

        protected static ITaskItem CreateTaskItem(string itemSpec, IDictionary metadata)
        {
            var meta = AdjustMetadata(metadata);
            return meta == null ? new TaskItem(itemSpec) : new TaskItem(itemSpec, meta);
        }

        private static IDictionary AdjustMetadata(IDictionary metadata)
        {
            if (metadata == null)
            {
                return null;
            }

            if (metadata is IDictionary<string, string>)
            {
                return metadata;
            }

            return metadata.Keys.Cast<object>()
                .ToDictionary(key => key.ToString(),
                    key =>
                        metadata[key] == null
                            ? string.Empty
                            : metadata[key] is string ? (string)metadata[key] : metadata[key].ToString());
        }

        private sealed class DebugLogManager : IHttpExchangeLoggerManager
        {
            private readonly Action<string> logHandler;

            public DebugLogManager(Action<string> logHandler)
            {
                this.logHandler = logHandler;
            }

            public IHttpExchangeLogger GetExchangeLogger() => new DebugLogger(logHandler);
        }

        private sealed class DebugLogger : IHttpExchangeLogger
        {
            private readonly Action<string> logHandler;

            public DebugLogger(Action<string> logHandler)
            {
                this.logHandler = logHandler ?? throw new ArgumentNullException(nameof(logHandler));
            }

            public void ExchangeEnd() => logHandler("RequestEnd");

            public void LogRequest(string request) => logHandler($"Request: {request}");

            public void LogRequestContent(string requestContent) => logHandler($"Request Content: {requestContent}");

            public void LogResponse(string response) => logHandler($"Response: {response}");

            public void LogResponseContent(string responseContent) => logHandler($"Response Content: {responseContent}");
        }
    }

    public sealed class TranslationServiceProject : SmartCatBaseTask
    {
        /// <summary>
        /// Идентификатор проекта в системе SmartCAT.
        /// Указывается либо ProjectId, либо ProjectName.
        /// ProjectId имеет приоритет перед ProjectName.
        /// </summary>
        public Guid ProjectId { get; set; } = Guid.Empty;

        /// <summary>
        /// Имя проекта в системе SmartCAT
        /// Указывается либо ProjectId, либо ProjectName.
        /// ProjectId имеет приоритет перед ProjectName.
        /// </summary>
        public string ProjectName { get; set; }

        /// <summary>
        /// Разрешает создавать новый или обновлять параметры существующего проекта (например, добавлять языки)
        /// </summary>
        public bool UpdateOrCreate { get; set; } = false;

        /// <summary>
        /// Краткое описание проекта
        /// </summary>
        public string Description { get; set; }

        /// <summary>
        /// Исходный язык проекта (имя языковой культуры в формате: ru, ru-Ru, en, en-GB и т.п.)
        /// </summary>
        public string SourceLanguage { get; set; }

        /// <summary>
        /// языки для перевода в проекте (задаются в формате имени культуры: ru, ru-Ru, en, en-GB, vi-VN и т.п.)
        /// </summary>
        public ITaskItem[] TargetLanguages { get; set; }

        /// <summary>
        /// Список идентификаторов документов в указанном проекте.
        /// </summary>
        [Output]
        public ITaskItem[] Documents { get; private set; }

        /// <summary>
        /// Все метаданные заданного проекта. Возвращается список из одного проекта если его имя или идентификатор найдены в системе SmartCAT
        /// </summary>
        [Output]
        public ITaskItem[] Project { get; private set; }

        protected override void ValidateParameters()
        {
            if (ProjectId == Guid.Empty && string.IsNullOrWhiteSpace(ProjectName))
            {
                throw new InvalidOperationException("Invalid pareameter");
            }
        }

        protected override void InternalExecute()
        {
            var project = InternalGetProject().Result;

            if (project == null)
            {
                var param = ProjectId != Guid.Empty
                    ? new object[] { "id", ProjectId }
                    : new object[] { "name", ProjectName };
                Log.LogMessage(MessageImportance.Normal, "Project with {0} '{1}' is not found.", param);
                return;
            }

            Project = new[] { CreateTaskItem(project.Id.ToString("B"), GetProjectMetadata(project)) };

            Documents =
                project.Documents.Select(model => CreateTaskItem(model.Id, GetDocumentMetadata(model)))
                    .ToArray();

        }

        private async Task<ProjectModel> GetProject(string projectName, string sourceLanguage = null)
        {
            if (projectName == null)
            {
                throw new ArgumentNullException(nameof(projectName));
            }

            var projects = await TranslationApiClient.Projects.List();
            return projects.SingleOrDefault(model => model.Name == projectName && (string.IsNullOrEmpty(sourceLanguage) || model.SourceLanguage.Equals(SourceLanguage)));
        }

        private async Task<ProjectModel> UpdateProject(ProjectModel project, ProjectChangesModel changes,
            IEnumerable<string> languagesToAdd)
        {
            if (project == null)
            {
                throw new ArgumentNullException(nameof(project));
            }

            var refresh = false;
            if (changes != null)
            {
                refresh = true;
                await TranslationApiClient.Projects.UpdateProject(project.Id, changes);
            }

            foreach (var lang in languagesToAdd.Except(project.TargetLanguages))
            {
                refresh = true;
                await TranslationApiClient.Projects.AddLanguage(project.Id, lang);
            }

            return refresh ? await TranslationApiClient.Projects.GetProject(project.Id) : project;
        }

        private async Task<ProjectModel> CreateProject(CreateProjectModel model)
        {
            if (model == null) throw new ArgumentNullException(nameof(model));

            return await TranslationApiClient.Projects.CreateProject(model, null);
        }

        private async Task<ProjectModel> InternalGetProject()
        {
            var result = await GetProject(ProjectId) ?? await GetProject(ProjectName, SourceLanguage);

            if (!UpdateOrCreate)
            {
                return result;
            }

            return await (result == null ? CreateProject(GetCreateContext()) : UpdateProject(result, GetChangeContext(result), TargetLanguages.Select(item => item.ItemSpec)));
        }

        private ProjectChangesModel GetChangeContext(ProjectModel project)
        {
            var changes = new Lazy<ProjectChangesModel>();

            if (!string.IsNullOrWhiteSpace(ProjectName) && !project.Name.Equals(ProjectName))
            {
                changes.Value.Name = ProjectName;
            }
            if (!string.IsNullOrWhiteSpace(Description) && !project.Description.Equals(Description))
            {
                changes.Value.Description = Description;
            }

            return changes.IsValueCreated ? changes.Value : null;
        }

        private CreateProjectModel GetCreateContext()
        {
            if (string.IsNullOrEmpty(ProjectName))
            {
                throw new InvalidOperationException("Undefined ProjectName");
            }

            return new CreateProjectModel
            {
                Name = ProjectName,
                Description = Description,
                SourceLanguage = SourceLanguage ?? "ru",
                TargetLanguages = TargetLanguages?.Select(item => item.ItemSpec) ?? new []{"en"},
                TranslationMemoryName = string.Empty,
                Pretranslate = true
            };
        }
    }

    public sealed class TranslationServiceUpload : SmartCatBaseTask
    {
        /// <summary>
        /// Идентификатор проекта в системе SmartCAT.
        /// </summary>
        [Required]
        public string ProjectId { get; set; }

        /// <summary>
        /// Исходные файлы (документы) для отправки (обновления) на сервис перевода
        /// </summary>
        [Required]
        public ITaskItem[] SourceDocuments { get; set; }

        /// <summary>
        /// Имя файла как оно будет отображаться в сервисе перевода (элементы в списке должны соотвествовать "1 к 1" списку SourceDocuments)
        /// </summary>
        public ITaskItem[] Destinations { get; set; }

        /// <summary>
        /// Включение и отключение разметки плейсхолдеров в переводах
        /// </summary>
        public bool EnablePlaceholders { get; set; } = false;

        /// <summary>
        /// Список идентификаторов загруженных или обновлённых документов в сервисе переводов
        /// </summary>
        [Output]
        public ITaskItem[] Documents { get; private set; }

        protected override void ValidateParameters()
        {
            if (Destinations == null)
            {
                Destinations = SourceDocuments.Select(item => new TaskItem(item.GetMetadata("DocumentName") ?? item.GetMetadata("FileName"))).ToArray();
            }
        }

        protected override void InternalExecute()
        {
            var projectId = new Guid(ProjectId);

            var project = GetProject(projectId).Result;

            if (project == null)
            {
                throw new InvalidOperationException("Project not found.");
            }

            var resourceMap = Enumerable.Range(0, SourceDocuments.Length).ToDictionary(i => SourceDocuments[i].GetFullPath(), i => Destinations[i].ItemSpec);

            Documents = new ITaskItem[] { };

            var documents = new List<ITaskItem>();

            foreach (var filename in resourceMap.Keys)
            {
                if (!File.Exists(filename))
                {
                    Log.LogMessage($"File '{filename}' is not found. Skiping...");
                    continue;
                }

                var resourceName = resourceMap[filename];

                var additionalMeta = new Dictionary<string, object>
                {
                    {"ProjectId", project.Id},
                    {"SourceDocument", filename}                
                };

                Log.LogMessage($"Uploading '{filename}' -> {project.Name}: {resourceName}");

                var result = InternalUpload(project, filename, resourceName, EnablePlaceholders)
                    .Result
                    .Select(model => CreateTaskItem(model.Id, GetDocumentMetadata(model).Join(additionalMeta)));

                Log.LogMessage("Upload successful.");

                documents.AddRange(result);
            }

            Documents = documents.ToArray();
        }

        public async Task<IReadOnlyCollection<DocumentModel>> InternalUpload(ProjectModel project, string documentPath, string targetName, bool enablePlaceholders = false)
        {
            using (var content = File.OpenRead(documentPath))
            {
                var file = new FileModel
                {
                    Content = content,
                    Name = targetName + ".csv",
                    MimeType = @"text/csv"
                };

                var document = project.Documents.FirstOrDefault(model => model.Name == targetName);
                return
                    (document == null
                        ? await
                            TranslationApiClient.Projects.AddDocument(project.Id, new[] {file},
                                new[] {new CreateDocumentPropertyModel {EnablePlaceholders = enablePlaceholders}})
                        : await
                            TranslationApiClient.Documents.UpdateDocument(document.Id, file,
                                new UploadDocumentPropertiesModel {EnablePlaceholders = enablePlaceholders}));
            }

        }
    }

    public sealed class TranslationServiceDownload : SmartCatBaseTask
    {
        /// <summary>
        /// Идентификатор проекта в системе SmartCAT.
        /// </summary>
        [Required]
        public string ProjectId { get; set; }

        /// <summary>
        /// Список идентификаторов документов в проекте, которые требуется скачать
        /// </summary>
        [Required]
        public ITaskItem[] Documents { get; set; }

        /// <summary>
        /// Список языков, переводы на которые нужно скачать
        /// </summary>
        public ITaskItem[] TargetLanguages { get; set; }

        /// <summary>
        /// Пути к файлам, куда нужно сохранить полученные переводы. Задаётся в виде списка путей.
        /// Конечное имя файла получается подстановкой имени культуры в имя файла из этого списка.
        /// </summary>
        [Required]
        public ITaskItem[] DestinationFiles { get; set; }

        /// <summary>
        /// Перезаписывать или нет файл на диске при сохранении переводов
        /// </summary>
        public bool Overwrite { get; set; }

        /// <summary>
        /// Список путей к файлам с полученными переводами.
        /// В метаданных указаны имена исходных документов в проекте на сервисе переводов и язык перевода.
        /// </summary>
        [Output]
        public ITaskItem[] Result { get; private set; }

        protected override void ValidateParameters()
        {
            if (Documents == null || Documents.Length == 0)
            {
                throw new ArgumentException(nameof(Documents));
            }
            if (DestinationFiles == null || DestinationFiles.Length == 0)
            {
                throw new ArgumentException(nameof(DestinationFiles));
            }
            if (Documents.Length != DestinationFiles.Length)
            {
                throw new ArgumentException("Incorrect destinations specified.");
            }
        }

        protected override void InternalExecute()
        {
            var projectId = new Guid(ProjectId);

            var project = GetProject(projectId).Result;

            if (project == null)
            {
                throw new InvalidOperationException("Project not found.");
            }

            // ключем карты маппинга культур должны быть имена из SmartCAT-проекта. 
            // Дополнительно фильтруем список языков в SmartCAT-проекте по опциональному списку из параметра TargetLanguages задачи.
            var languageMap = project.TargetLanguages
                .Select(item => new KeyValuePair<string, CultureInfo>(item, CultureInfo.CreateSpecificCulture(item)))
                .Where(
                    pair =>
                        TargetLanguages == null ||
                        TargetLanguages.Any(item => item.ItemSpec.Equals(pair.Value.Name, StringComparison.InvariantCultureIgnoreCase)))
                .ToDictionary(pair => pair.Key, pair => pair.Value);

            Log.LogMessage("Loading translations from '{0}' project for languages: {1}", project.Name, string.Join(",", languageMap.Keys));

            var file = InternalDownload(project, Documents.Select(item => item.ItemSpec), languageMap.Keys);

            if (file == null)
            {
                Log.LogWarning("Requested translations are not found...");
                return;
            }

            var resourceMap = Enumerable.Range(0, Documents.Length)
                .ToDictionary(i => Documents[i].ItemSpec, i => DestinationFiles[i].ItemSpec);
            string PathMapper(string resource) => resourceMap[resource];

            var documents = (file.MimeType == "application/zip")
                ? ExtractTranslations(file, PathMapper, languageMap, Overwrite)
                : StorePureTranslation(file, PathMapper, languageMap, Overwrite);

            Result = documents.Select(tuple => CreateTaskItem(tuple.Item1, new Dictionary<string, string>
            {
                {"Resource", tuple.Item2.ResourceName},
                {"Language", tuple.Item2.TranslationLanguage},
                {"Culture", languageMap[tuple.Item2.TranslationLanguage].Name}
            })).ToArray();
        }

        private FileModel InternalDownload(ProjectModel project, IEnumerable<string> requestedResources, IEnumerable<string> targetTranslations)
        {
            if (requestedResources == null)
            {
                throw new ArgumentNullException(nameof(requestedResources));
            };

            var resources = new HashSet<string>(requestedResources);
            var languages = (targetTranslations ?? project.TargetLanguages).ToList();
            var documents = project.Documents.Where(model => resources.Contains(model.Name) && languages.Contains(model.TargetLanguage)).ToList();

            var fixedFileName = documents.Count == 1 && languages.Count == 1 ? $"{documents.First().Name}({languages.First()}).csv" : null;

            if (documents.Count == 0)
            {
                return null;
            }

            return TranslationApiClient.Documents.RequestExport(documents.Select(model => model.Id))
                .ContinueWith(
                    task =>
                    {
                        var file = DoPollingRequestAsync(task.Result,
                            model => TranslationApiClient.Documents.DownloadExportResult(model.Id)).Result;

                        if (file != null && file.MimeType == "text/csv")
                        {
                            file.Name = file.Name ?? fixedFileName;
                        }

                        return file;
                    }
                    ,
                    TaskContinuationOptions.LongRunning |
                    TaskContinuationOptions.NotOnCanceled |
                    TaskContinuationOptions.NotOnFaulted)
                .Result;
        }

        private readonly TimeSpan pollingRequestRequeryInterval = TimeSpan.FromSeconds(2);

        private async Task<TResult> DoPollingRequestAsync<TModel, TResult>(TModel model, Func<TModel, Task<TResult>> action)
        {
            if (action == null) throw new ArgumentNullException(nameof(action));

            var result = await action(model);
            while (result == null)
            {
                await Task.Delay(pollingRequestRequeryInterval);
                result = await action(model);
            }

            return result;
        }

        private string ResolveResourceTargetPath(ExportedResourceParsedName entry, Func<string, string> targetPathResourceResolver, IDictionary<string, CultureInfo> languageMap)
        {
            var targetpattern = targetPathResourceResolver(entry.ResourceName);
            var targetPath = entry.GetTargetFileName(targetpattern, languageMap, entry.Extension);
            var destinationDir = Path.GetDirectoryName(targetPath);
            if (destinationDir != null && !Directory.Exists(destinationDir))
            {
                Directory.CreateDirectory(destinationDir);
            }
            return targetPath;
        }

        private IEnumerable<Tuple<string, ExportedResourceParsedName>> StorePureTranslation(FileModel file, Func<string, string> mapTargetPath, IDictionary<string, CultureInfo> languageMap, bool overwrite)
        {
            var parsedName = ExportedResourceParsedName.Parse(file.Name);
            var path = ResolveResourceTargetPath(parsedName, mapTargetPath, languageMap);

            using (var targetFile = File.Open(path, overwrite ? FileMode.Create : FileMode.CreateNew))
            {
                file.Content.CopyTo(targetFile);
            }

            Log.LogMessage("Downloaded resource '{0}' saved to '{1}'", file.Name, path);

            return new[] { new Tuple<string, ExportedResourceParsedName>(path, parsedName) };
        }

        private IEnumerable<Tuple<string, ExportedResourceParsedName>> ExtractTranslations(FileModel archive, Func<string, string> mapTargetPath, IDictionary<string, CultureInfo> languageMap, bool overwrite)
        {
            var files = new List<Tuple<string, ExportedResourceParsedName>>();

            using (var zip = new ZipArchive(archive.Content))
            {
                foreach (var entry in zip.Entries)
                {
                    var parsedName = ExportedResourceParsedName.Parse(entry.Name);
                    var path = ResolveResourceTargetPath(parsedName, mapTargetPath, languageMap);

                    entry.ExtractToFile(path, overwrite);

                    Log.LogMessage("Downloaded resource '{0}' extracted to '{1}'", entry.FullName, path);

                    files.Add(new Tuple<string, ExportedResourceParsedName>(path, parsedName));
                }
            }

            return files;
        }

        private sealed class ExportedResourceParsedName
        {
            private static readonly Regex MatchLangFromFileName = new Regex(@"^\""?(?<name>.*)\((?<lang>[a-zA-z.]{2}(?:-[a-zA-Z]+)?)\)(?<extension>\.\w*)?\""?", RegexOptions.IgnoreCase | RegexOptions.Compiled | RegexOptions.Singleline | RegexOptions.CultureInvariant);

            private ExportedResourceParsedName(string name) => InternalParse(name ?? throw new ArgumentNullException(nameof(name)));

            private void InternalParse(string name)
            {
                var match = MatchLangFromFileName.Match(name);

                if (!match.Success)
                {
                    throw new InvalidOperationException($"Exported resource name parsing error: {name}");
                }

                ResourceName = match.Groups["name"].Value;
                TranslationLanguage = match.Groups["lang"].Value;
                Extension = match.Groups["extension"].Value;
            }

            internal static ExportedResourceParsedName Parse(string name) => new ExportedResourceParsedName(name);

            internal string GetTargetFileName(string originalPath, IDictionary<string, CultureInfo> languageMap, string extension = null)
            {
                var name =
                    $"{Path.GetFileNameWithoutExtension(originalPath)}.{languageMap[TranslationLanguage].Name}{extension ?? Extension}";

                return Path.Combine(Path.GetDirectoryName(originalPath) ?? string.Empty, name);
            }

            public string Extension { get; private set; }

            public string TranslationLanguage { get; private set; }

            public string ResourceName { get; private set; }
        }
    }
    
    internal static class TaskItemExtensions
    {
        public static string GetFullPath(this ITaskItem item) => item.GetMetadata("FullPath");

        public static string GetFileName(this ITaskItem item) => item.GetMetadata("Filename");

        public static string GetExtension(this ITaskItem item) => item.GetMetadata("Extension");

        internal static string GetName(this ITaskItem item) => $"{item.GetFileName()}{item.GetExtension()}";
    }

    internal static class MetadataExtension
    {
        internal static IDictionary Join(this IDictionary primary, IDictionary scondary)
        {
            if (primary == null) throw new ArgumentNullException(nameof(primary));
            if (scondary == null) throw new ArgumentNullException(nameof(scondary));

            return new[] { primary, scondary }
                .SelectMany(dictionary => dictionary.Keys.Cast<object>().Select(o => new KeyValuePair<object, object>(o, dictionary[o])))
                .ToLookup(pair => pair.Key, pair => pair.Value)
                .ToDictionary(group => group.Key, group => group.First());
        }

    }
}
